"""Tests for hermite_e module."""

from functools import reduce

import numpy as np
from jax import config
from numpy.polynomial.polynomial import polyval
from numpy.testing import (
    assert_,
    assert_array_almost_equal,
    assert_array_equal,
    assert_raises,
)

import orthax.hermite_e as herme

config.update("jax_enable_x64", True)

He0 = np.array([1])
He1 = np.array([0, 1])
He2 = np.array([-1, 0, 1])
He3 = np.array([0, -3, 0, 1])
He4 = np.array([3, 0, -6, 0, 1])
He5 = np.array([0, 15, 0, -10, 0, 1])
He6 = np.array([-15, 0, 45, 0, -15, 0, 1])
He7 = np.array([0, -105, 0, 105, 0, -21, 0, 1])
He8 = np.array([105, 0, -420, 0, 210, 0, -28, 0, 1])
He9 = np.array([0, 945, 0, -1260, 0, 378, 0, -36, 0, 1])

Helist = [He0, He1, He2, He3, He4, He5, He6, He7, He8, He9]


def trim(x):
    return herme.hermetrim(x, tol=1e-6)


class TestConstants:
    def test_hermedomain(self):
        assert_array_equal(herme.hermedomain, [-1, 1])

    def test_hermezero(self):
        assert_array_equal(herme.hermezero, [0])

    def test_hermeone(self):
        assert_array_equal(herme.hermeone, [1])

    def test_hermex(self):
        assert_array_equal(herme.hermex, [0, 1])


class TestArithmetic:
    x = np.linspace(-3, 3, 100)

    def test_hermeadd(self):
        for i in range(5):
            for j in range(5):
                msg = f"At i={i}, j={j}"
                tgt = np.zeros(max(i, j) + 1)
                tgt[i] += 1
                tgt[j] += 1
                res = herme.hermeadd([0] * i + [1], [0] * j + [1])
                assert_array_equal(trim(res), trim(tgt), err_msg=msg)

    def test_hermesub(self):
        for i in range(5):
            for j in range(5):
                msg = f"At i={i}, j={j}"
                tgt = np.zeros(max(i, j) + 1)
                tgt[i] += 1
                tgt[j] -= 1
                res = herme.hermesub([0] * i + [1], [0] * j + [1])
                assert_array_equal(trim(res), trim(tgt), err_msg=msg)

    def test_hermemulx(self):
        assert_array_equal(trim(herme.hermemulx([0])), [0])
        assert_array_equal(trim(herme.hermemulx([1])), [0, 1])
        for i in range(1, 5):
            ser = [0] * i + [1]
            tgt = [0] * (i - 1) + [i, 0, 1]
            assert_array_equal(trim(herme.hermemulx(ser)), tgt)

    def test_hermemul(self):
        # check values of result
        for i in range(5):
            pol1 = [0] * i + [1]
            val1 = herme.hermeval(self.x, pol1)
            for j in range(5):
                msg = f"At i={i}, j={j}"
                pol2 = [0] * j + [1]
                val2 = herme.hermeval(self.x, pol2)
                pol3 = herme.hermemul(pol1, pol2)
                val3 = herme.hermeval(self.x, pol3)
                assert_(len(pol3) == i + j + 1, msg)
                assert_array_almost_equal(val3, val1 * val2, err_msg=msg)

    def test_hermediv(self):
        for i in range(5):
            for j in range(5):
                msg = f"At i={i}, j={j}"
                ci = [0] * i + [1]
                cj = [0] * j + [1]
                tgt = herme.hermeadd(ci, cj)
                quo, rem = herme.hermediv(tgt, ci)
                res = herme.hermeadd(herme.hermemul(quo, ci), rem)
                assert_array_equal(trim(res), trim(tgt), err_msg=msg)

    def test_hermepow(self):
        for i in range(5):
            for j in range(5):
                msg = f"At i={i}, j={j}"
                c = np.arange(i + 1)
                tgt = reduce(herme.hermemul, [c] * j, np.array([1]))
                res = herme.hermepow(c, j)
                assert_array_equal(trim(res), trim(tgt), err_msg=msg)


class TestEvaluation:
    # coefficients of 1 + 2*x + 3*x**2
    c1d = np.array([4.0, 2.0, 3.0])
    c2d = np.einsum("i,j->ij", c1d, c1d)
    c3d = np.einsum("i,j,k->ijk", c1d, c1d, c1d)

    # some random values in [-1, 1)
    x = np.random.random((3, 5)) * 2 - 1
    y = polyval(x, [1.0, 2.0, 3.0])

    def test_hermeval(self):
        # check empty input
        assert_array_equal(herme.hermeval([], [1]).size, 0)

        # check normal input)
        x = np.linspace(-1, 1)
        y = [polyval(x, c) for c in Helist]
        for i in range(10):
            msg = f"At i={i}"
            tgt = y[i]
            res = herme.hermeval(x, [0] * i + [1])
            assert_array_almost_equal(res, tgt, err_msg=msg)

        # check that shape is preserved
        for i in range(3):
            dims = [2] * i
            x = np.zeros(dims)
            assert_array_equal(herme.hermeval(x, [1]).shape, dims)
            assert_array_equal(herme.hermeval(x, [1, 0]).shape, dims)
            assert_array_equal(herme.hermeval(x, [1, 0, 0]).shape, dims)

    def test_hermeval2d(self):
        x1, x2, x3 = self.x
        y1, y2, y3 = self.y

        # test exceptions
        assert_raises(ValueError, herme.hermeval2d, x1, x2[:2], self.c2d)

        # test values
        tgt = y1 * y2
        res = herme.hermeval2d(x1, x2, self.c2d)
        assert_array_almost_equal(res, tgt)

        # test shape
        z = np.ones((2, 3))
        res = herme.hermeval2d(z, z, self.c2d)
        assert_(res.shape == (2, 3))

    def test_hermeval3d(self):
        x1, x2, x3 = self.x
        y1, y2, y3 = self.y

        # test exceptions
        assert_raises(ValueError, herme.hermeval3d, x1, x2, x3[:2], self.c3d)

        # test values
        tgt = y1 * y2 * y3
        res = herme.hermeval3d(x1, x2, x3, self.c3d)
        assert_array_almost_equal(res, tgt)

        # test shape
        z = np.ones((2, 3))
        res = herme.hermeval3d(z, z, z, self.c3d)
        assert_(res.shape == (2, 3))

    def test_hermegrid2d(self):
        x1, x2, x3 = self.x
        y1, y2, y3 = self.y

        # test values
        tgt = np.einsum("i,j->ij", y1, y2)
        res = herme.hermegrid2d(x1, x2, self.c2d)
        assert_array_almost_equal(res, tgt)

        # test shape
        z = np.ones((2, 3))
        res = herme.hermegrid2d(z, z, self.c2d)
        assert_(res.shape == (2, 3) * 2)

    def test_hermegrid3d(self):
        x1, x2, x3 = self.x
        y1, y2, y3 = self.y

        # test values
        tgt = np.einsum("i,j,k->ijk", y1, y2, y3)
        res = herme.hermegrid3d(x1, x2, x3, self.c3d)
        assert_array_almost_equal(res, tgt)

        # test shape
        z = np.ones((2, 3))
        res = herme.hermegrid3d(z, z, z, self.c3d)
        assert_(res.shape == (2, 3) * 3)


class TestIntegral:
    def test_hermeint(self):  # noqa:C901
        # check exceptions
        assert_raises(TypeError, herme.hermeint, [0], 0.5)
        assert_raises(ValueError, herme.hermeint, [0], -1)
        assert_raises(ValueError, herme.hermeint, [0], 1, [0, 0])
        assert_raises(ValueError, herme.hermeint, [0], lbnd=[0])
        assert_raises(ValueError, herme.hermeint, [0], scl=[0])
        assert_raises(TypeError, herme.hermeint, [0], axis=0.5)

        # test integration of zero polynomial
        for i in range(2, 5):
            k = [0] * (i - 2) + [1]
            res = herme.hermeint([0], m=i, k=k)
            assert_array_almost_equal(trim(res), [0, 1])

        # check single integration with integration constant
        for i in range(5):
            scl = i + 1
            pol = [0] * i + [1]
            tgt = [i] + [0] * i + [1 / scl]
            hermepol = herme.poly2herme(pol)
            hermeint = herme.hermeint(hermepol, m=1, k=[i])
            res = herme.herme2poly(hermeint)
            assert_array_almost_equal(trim(res), trim(tgt))

        # check single integration with integration constant and lbnd
        for i in range(5):
            scl = i + 1
            pol = [0] * i + [1]
            hermepol = herme.poly2herme(pol)
            hermeint = herme.hermeint(hermepol, m=1, k=[i], lbnd=-1)
            assert_array_almost_equal(herme.hermeval(-1, hermeint), i)

        # check single integration with integration constant and scaling
        for i in range(5):
            scl = i + 1
            pol = [0] * i + [1]
            tgt = [i] + [0] * i + [2 / scl]
            hermepol = herme.poly2herme(pol)
            hermeint = herme.hermeint(hermepol, m=1, k=[i], scl=2)
            res = herme.herme2poly(hermeint)
            assert_array_almost_equal(trim(res), trim(tgt))

        # check multiple integrations with default k
        for i in range(5):
            for j in range(2, 5):
                pol = [0] * i + [1]
                tgt = pol[:]
                for k in range(j):
                    tgt = herme.hermeint(tgt, m=1)
                res = herme.hermeint(pol, m=j)
                assert_array_almost_equal(trim(res), trim(tgt))

        # check multiple integrations with defined k
        for i in range(5):
            for j in range(2, 5):
                pol = [0] * i + [1]
                tgt = pol[:]
                for k in range(j):
                    tgt = herme.hermeint(tgt, m=1, k=[k])
                res = herme.hermeint(pol, m=j, k=list(range(j)))
                assert_array_almost_equal(trim(res), trim(tgt))

        # check multiple integrations with lbnd
        for i in range(5):
            for j in range(2, 5):
                pol = [0] * i + [1]
                tgt = pol[:]
                for k in range(j):
                    tgt = herme.hermeint(tgt, m=1, k=[k], lbnd=-1)
                res = herme.hermeint(pol, m=j, k=list(range(j)), lbnd=-1)
                assert_array_almost_equal(trim(res), trim(tgt))

        # check multiple integrations with scaling
        for i in range(5):
            for j in range(2, 5):
                pol = [0] * i + [1]
                tgt = pol[:]
                for k in range(j):
                    tgt = herme.hermeint(tgt, m=1, k=[k], scl=2)
                res = herme.hermeint(pol, m=j, k=list(range(j)), scl=2)
                assert_array_almost_equal(trim(res), trim(tgt))

    def test_hermeint_axis(self):
        # check that axis keyword works
        c2d = np.random.random((3, 4))

        tgt = np.vstack([herme.hermeint(c) for c in c2d.T]).T
        res = herme.hermeint(c2d, axis=0)
        assert_array_almost_equal(res, tgt)

        tgt = np.vstack([herme.hermeint(c) for c in c2d])
        res = herme.hermeint(c2d, axis=1)
        assert_array_almost_equal(res, tgt)

        tgt = np.vstack([herme.hermeint(c, k=3) for c in c2d])
        res = herme.hermeint(c2d, k=3, axis=1)
        assert_array_almost_equal(res, tgt)


class TestDerivative:
    def test_hermeder(self):
        # check exceptions
        assert_raises(TypeError, herme.hermeder, [0], 0.5)
        assert_raises(ValueError, herme.hermeder, [0], -1)

        # check that zeroth derivative does nothing
        for i in range(5):
            tgt = [0] * i + [1]
            res = herme.hermeder(tgt, m=0)
            assert_array_equal(trim(res), trim(tgt))

        # check that derivation is the inverse of integration
        for i in range(5):
            for j in range(2, 5):
                tgt = [0] * i + [1]
                res = herme.hermeder(herme.hermeint(tgt, m=j), m=j)
                assert_array_almost_equal(trim(res), trim(tgt))

        # check derivation with scaling
        for i in range(5):
            for j in range(2, 5):
                tgt = [0] * i + [1]
                res = herme.hermeder(herme.hermeint(tgt, m=j, scl=2), m=j, scl=0.5)
                assert_array_almost_equal(trim(res), trim(tgt))

    def test_hermeder_axis(self):
        # check that axis keyword works
        c2d = np.random.random((3, 4))

        tgt = np.vstack([herme.hermeder(c) for c in c2d.T]).T
        res = herme.hermeder(c2d, axis=0)
        assert_array_almost_equal(res, tgt)

        tgt = np.vstack([herme.hermeder(c) for c in c2d])
        res = herme.hermeder(c2d, axis=1)
        assert_array_almost_equal(res, tgt)


class TestVander:
    # some random values in [-1, 1)
    x = np.random.random((3, 5)) * 2 - 1

    def test_hermevander(self):
        # check for 1d x
        x = np.arange(3)
        v = herme.hermevander(x, 3)
        assert_(v.shape == (3, 4))
        for i in range(4):
            coef = [0] * i + [1]
            assert_array_almost_equal(v[..., i], herme.hermeval(x, coef))

        # check for 2d x
        x = np.array([[1, 2], [3, 4], [5, 6]])
        v = herme.hermevander(x, 3)
        assert_(v.shape == (3, 2, 4))
        for i in range(4):
            coef = [0] * i + [1]
            assert_array_almost_equal(v[..., i], herme.hermeval(x, coef))

    def test_hermevander2d(self):
        # also tests hermeval2d for non-square coefficient array
        x1, x2, x3 = self.x
        c = np.random.random((2, 3))
        van = herme.hermevander2d(x1, x2, (1, 2))
        tgt = herme.hermeval2d(x1, x2, c)
        res = np.dot(van, c.flat)
        assert_array_almost_equal(res, tgt)

        # check shape
        van = herme.hermevander2d([x1], [x2], (1, 2))
        assert_(van.shape == (1, 5, 6))

    def test_hermevander3d(self):
        # also tests hermeval3d for non-square coefficient array
        x1, x2, x3 = self.x
        c = np.random.random((2, 3, 4))
        van = herme.hermevander3d(x1, x2, x3, (1, 2, 3))
        tgt = herme.hermeval3d(x1, x2, x3, c)
        res = np.dot(van, c.flat)
        assert_array_almost_equal(res, tgt)

        # check shape
        van = herme.hermevander3d([x1], [x2], [x3], (1, 2, 3))
        assert_(van.shape == (1, 5, 24))


class TestFitting:
    def test_hermefit(self):
        def f(x):
            return x * (x - 1) * (x - 2)

        def f2(x):
            return x**4 + x**2 + 1

        # Test exceptions
        assert_raises(ValueError, herme.hermefit, [1], [1], -1)
        assert_raises(TypeError, herme.hermefit, [[1]], [1], 0)
        assert_raises(TypeError, herme.hermefit, [], [1], 0)
        assert_raises(TypeError, herme.hermefit, [1], [[[1]]], 0)
        assert_raises(TypeError, herme.hermefit, [1, 2], [1], 0)
        assert_raises(TypeError, herme.hermefit, [1], [1, 2], 0)
        assert_raises(TypeError, herme.hermefit, [1], [1], 0, w=[[1]])
        assert_raises(TypeError, herme.hermefit, [1], [1], 0, w=[1, 1])
        assert_raises(ValueError, herme.hermefit, [1], [1], (-1,))
        assert_raises(ValueError, herme.hermefit, [1], [1], (2, -1, 6))
        assert_raises(TypeError, herme.hermefit, [1], [1], ())

        # Test fit
        x = np.linspace(0, 2)
        y = f(x)
        #
        coef3 = herme.hermefit(x, y, 3)
        assert_array_equal(len(coef3), 4)
        assert_array_almost_equal(herme.hermeval(x, coef3), y)
        coef3 = herme.hermefit(x, y, (0, 1, 2, 3))
        assert_array_equal(len(coef3), 4)
        assert_array_almost_equal(herme.hermeval(x, coef3), y)
        #
        coef4 = herme.hermefit(x, y, 4)
        assert_array_equal(len(coef4), 5)
        assert_array_almost_equal(herme.hermeval(x, coef4), y)
        coef4 = herme.hermefit(x, y, (0, 1, 2, 3, 4))
        assert_array_equal(len(coef4), 5)
        assert_array_almost_equal(herme.hermeval(x, coef4), y)
        # check things still work if deg is not in strict increasing
        coef4 = herme.hermefit(x, y, (2, 3, 4, 1, 0))
        assert_array_equal(len(coef4), 5)
        assert_array_almost_equal(herme.hermeval(x, coef4), y)
        #
        coef2d = herme.hermefit(x, np.array([y, y]).T, 3)
        assert_array_almost_equal(coef2d, np.array([coef3, coef3]).T)
        coef2d = herme.hermefit(x, np.array([y, y]).T, (0, 1, 2, 3))
        assert_array_almost_equal(coef2d, np.array([coef3, coef3]).T)
        # test weighting
        w = np.zeros_like(x)
        yw = y.copy()
        w[1::2] = 1
        y[0::2] = 0
        wcoef3 = herme.hermefit(x, yw, 3, w=w)
        assert_array_almost_equal(wcoef3, coef3)
        wcoef3 = herme.hermefit(x, yw, (0, 1, 2, 3), w=w)
        assert_array_almost_equal(wcoef3, coef3)
        #
        wcoef2d = herme.hermefit(x, np.array([yw, yw]).T, 3, w=w)
        assert_array_almost_equal(wcoef2d, np.array([coef3, coef3]).T)
        wcoef2d = herme.hermefit(x, np.array([yw, yw]).T, (0, 1, 2, 3), w=w)
        assert_array_almost_equal(wcoef2d, np.array([coef3, coef3]).T)
        # test scaling with complex values x points whose square
        # is zero when summed.
        x = [1, 1j, -1, -1j]
        assert_array_almost_equal(herme.hermefit(x, x, 1), [0, 1])
        assert_array_almost_equal(herme.hermefit(x, x, (0, 1)), [0, 1])
        # test fitting only even polynomials
        x = np.linspace(-1, 1)
        y = f2(x)
        coef1 = herme.hermefit(x, y, 4)
        assert_array_almost_equal(herme.hermeval(x, coef1), y)
        coef2 = herme.hermefit(x, y, (0, 2, 4))
        assert_array_almost_equal(herme.hermeval(x, coef2), y)
        assert_array_almost_equal(coef1, coef2)


class TestCompanion:
    def test_raises(self):
        assert_raises(ValueError, herme.hermecompanion, [])
        assert_raises(ValueError, herme.hermecompanion, [1])

    def test_dimensions(self):
        for i in range(1, 5):
            coef = [0] * i + [1]
            assert_(herme.hermecompanion(coef).shape == (i, i))

    def test_linear_root(self):
        assert_(herme.hermecompanion([1, 2])[0, 0] == -0.5)


class TestGauss:
    def test_100(self):
        x, w = herme.hermegauss(100)

        # test orthogonality. Note that the results need to be normalized,
        # otherwise the huge values that can arise from fast growing
        # functions like Laguerre can be very confusing.
        v = herme.hermevander(x, 99)
        vv = np.dot(v.T * w, v)
        vd = 1 / np.sqrt(vv.diagonal())
        vv = vd[:, None] * vv * vd
        assert_array_almost_equal(vv, np.eye(100))

        # check that the integral of 1 is correct
        tgt = np.sqrt(2 * np.pi)
        assert_array_almost_equal(w.sum(), tgt)


class TestMisc:
    def test_hermefromroots(self):
        res = herme.hermefromroots([])
        assert_array_almost_equal(trim(res), [1])
        for i in range(1, 5):
            roots = np.cos(np.linspace(-np.pi, 0, 2 * i + 1)[1::2])
            pol = herme.hermefromroots(roots)
            res = herme.hermeval(roots, pol)
            tgt = 0
            assert_(len(pol) == i + 1)
            assert_array_almost_equal(herme.herme2poly(pol)[-1], 1)
            assert_array_almost_equal(res, tgt)

    def test_hermeroots(self):
        assert_array_almost_equal(herme.hermeroots([1]), [])
        assert_array_almost_equal(herme.hermeroots([1, 1]), [-1])
        for i in range(2, 5):
            tgt = np.linspace(-1, 1, i)
            res = herme.hermeroots(herme.hermefromroots(tgt))
            assert_array_almost_equal(trim(res), trim(tgt))

    def test_hermetrim(self):
        coef = [2, -1, 1, 0]

        # Test exceptions
        assert_raises(ValueError, herme.hermetrim, coef, -1)

        # Test results
        assert_array_equal(herme.hermetrim(coef), coef[:-1])
        assert_array_equal(herme.hermetrim(coef, 1), coef[:-3])
        assert_array_equal(herme.hermetrim(coef, 2), [0])

    def test_hermeline(self):
        assert_array_equal(herme.hermeline(3, 4), [3, 4])

    def test_herme2poly(self):
        for i in range(10):
            assert_array_almost_equal(herme.herme2poly([0] * i + [1]), Helist[i])

    def test_poly2herme(self):
        for i in range(10):
            assert_array_almost_equal(herme.poly2herme(Helist[i]), [0] * i + [1])

    def test_weight(self):
        x = np.linspace(-5, 5, 11)
        tgt = np.exp(-0.5 * x**2)
        res = herme.hermeweight(x)
        assert_array_almost_equal(res, tgt)



"""Tests for general orthogonal series functions."""

import numpy as np
import pytest
from jax import config
from numpy.testing import assert_allclose

import orthax

config.update("jax_enable_x64", True)

names = [
    "leg",
    "sleg",
    "chebT",
    "chebU",
    "chebV",
    "chebW",
    "geg",
    "jac",
    "lag",
    "glag",
    "herm",
    "herme",
]

recs = {
    "leg": (orthax.recurrence.Legendre, (), 1e-12),
    "sleg": (orthax.recurrence.ShiftedLegendre, (), 1e-12),
    "chebT": (orthax.recurrence.ChebyshevT, (), 1e-8),
    "chebU": (orthax.recurrence.ChebyshevU, (), 1e-8),
    "chebV": (orthax.recurrence.ChebyshevV, (), 1e-8),
    "chebW": (orthax.recurrence.ChebyshevW, (), 1e-8),
    "jac_pp": (orthax.recurrence.Jacobi, (0.27, 0.27), 1e-8),
    "jac_pn": (orthax.recurrence.Jacobi, (0.27, -0.27), 1e-8),
    "jac_np": (orthax.recurrence.Jacobi, (-0.27, 0.27), 1e-8),
    "jac_nn": (orthax.recurrence.Jacobi, (-0.27, -0.27), 1e-8),
    "geg_pos": (orthax.recurrence.Gegenbauer, (0.1,), 1e-8),
    "geg_neg": (orthax.recurrence.Gegenbauer, (-0.1,), 1e-6),
    "lag": (orthax.recurrence.Laguerre, (), 1e-12),
    "glag_pos": (orthax.recurrence.GeneralizedLaguerre, (2.5,), 1e-8),
    "glag_neg": (orthax.recurrence.GeneralizedLaguerre, (-0.34,), 1e-8),
    "herm": (orthax.recurrence.Hermite, (), 1e-12),
    "herme": (orthax.recurrence.HermiteE, (), 1e-12),
}


@pytest.mark.parametrize("scale", ["monic", "normalized"])
@pytest.mark.parametrize("name", recs.keys())
def test_generate_recurrence(name, scale):
    rec, args, tol = recs[name]
    rec1 = rec(*args, scale=scale)

    n = 10
    rec2 = orthax.recurrence.generate_recurrence(
        rec1.weight, rec1.domain, n, scale=scale
    )

    nn = np.arange(n)

    assert_allclose(rec1.a(nn), rec2.a(nn), atol=tol, rtol=tol)
    assert_allclose(rec1.b(nn), rec2.b(nn), atol=tol, rtol=tol)
    assert_allclose(rec1.g(nn), rec2.g(nn), atol=tol, rtol=tol)
    assert_allclose(abs(rec1.m(nn)), abs(rec2.m(nn)), atol=tol, rtol=tol)



"""
====================================
General Orthogonal Polynomial Series
====================================

The base ``orthax`` module provides a number of functions useful for dealing with
general orthogonal polynomial series, based around the idea of the three term
recurrence relation.

Arithmetic
----------

.. autosummary::
   :toctree: generated/

   orthadd
   orthsub
   orthmulx
   orthmul
   orthpow
   orthval
   orthval2d
   orthval3d
   orthgrid2d
   orthgrid3d

Calculus
--------

.. autosummary::
   :toctree: generated/

   orthder
   orthint

Misc Functions
--------------

.. autosummary::
   :toctree: generated/

   orthfromroots
   orthroots
   orthvander
   orthvander2d
   orthvander3d
   orthgauss
   orthweight
   orthnorm
   orthcompanion
   orthfit
   orthtrim
   orthline
   orth2poly
   poly2orth

"""

from . import (
    _version,
    chebyshev,
    hermite,
    hermite_e,
    laguerre,
    legendre,
    polynomial,
    polyutils,
    recurrence,
)
from ._general import (
    orth2poly,
    orthadd,
    orthcompanion,
    orthder,
    orthdiv,
    orthfit,
    orthfromroots,
    orthgauss,
    orthgrid2d,
    orthgrid3d,
    orthint,
    orthline,
    orthmul,
    orthmulx,
    orthnorm,
    orthpow,
    orthroots,
    orthsub,
    orthtrim,
    orthval,
    orthval2d,
    orthval3d,
    orthvander,
    orthvander2d,
    orthvander3d,
    orthweight,
    poly2orth,
)

__version__ = _version.get_versions()["version"]



"""
============================
Orthogonal Polynomial Series
============================

This module provides a number of functions useful for dealing with general orthogonal
polynomial series.

Arithmetic
----------

.. autosummary::
   :toctree: generated/

   orthadd
   orthsub
   orthmulx
   orthmul
   orthpow
   orthval
   orthval2d
   orthval3d
   orthgrid2d
   orthgrid3d

Calculus
--------

.. autosummary::
   :toctree: generated/

   orthder
   orthint

Misc Functions
--------------

.. autosummary::
   :toctree: generated/

   orthfromroots
   orthroots
   orthvander
   orthvander2d
   orthvander3d
   orthgauss
   orthweight
   orthnorm
   orthcompanion
   orthfit
   orthtrim
   orthline
   orth2poly
   poly2orth

"""

import functools

import jax
import jax.numpy as jnp
from jax import jit

from . import polyutils as pu

__all__ = [
    "orthadd",
    "orthsub",
    "orthmulx",
    "orthmul",
    "orthpow",
    "orthval",
    "orthval2d",
    "orthval3d",
    "orthgrid2d",
    "orthgrid3d",
    "orthder",
    "orthint",
    "orthfromroots",
    "orthroots",
    "orthvander",
    "orthvander2d",
    "orthvander3d",
    "orthgauss",
    "orthweight",
    "orthnorm",
    "orthcompanion",
    "orthfit",
    "orthtrim",
    "orthline",
    "orth2poly",
    "poly2orth",
]

orthtrim = pu.trimcoef


def tridiagmv(d, l, u, x):
    """Matvec for tridiagonal matrix."""
    a = u * x[1:]
    b = d * x
    c = l * x[:-1]
    return b.at[:-1].add(a).at[1:].add(c)


def last_nonzero(x):
    """Get index and value of last nonzero element of x"""
    i = len(x) - 1 - jnp.nonzero(x[::-1], size=1, fill_value=len(x))[0][0]
    return i, x[i]


@jit
def poly2orth(pol, rec):
    """Convert a polynomial to an orthogonal series.

    Convert an array representing the coefficients of a polynomial (relative
    to the "standard" basis) ordered from lowest degree to highest, to an
    array of the coefficients of the equivalent Chebyshev series, ordered
    from lowest to highest degree.

    Parameters
    ----------
    pol : array_like
        1-D array containing the polynomial coefficients
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    c : ndarray
        1-D array containing the coefficients of the equivalent orthogonal
        series.

    See Also
    --------
    orth2poly

    """
    pol = pu.as_series(pol)
    deg = len(pol) - 1

    res = jnp.zeros_like(pol)

    def body(i, res):
        k = deg - i
        res = orthadd(orthmulx(res, rec, mode="same"), pol[k], rec)
        return res

    res = jax.lax.fori_loop(0, deg + 1, body, res)
    return res


@jit
def orth2poly(c, rec):
    """Convert an orthogonal series to a polynomial in standard basis.

    Convert an array representing the coefficients of an orthogonal series,
    ordered from lowest degree to highest, to an array of the coefficients
    of the equivalent polynomial (relative to the "standard" basis) ordered
    from lowest to highest degree.

    Parameters
    ----------
    c : array_like
        1-D array containing the orthogonal series coefficients, ordered
        from lowest order term to highest.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    pol : ndarray
        1-D array containing the coefficients of the equivalent polynomial
        (relative to the "standard" basis) ordered from lowest order term
        to highest.

    See Also
    --------
    poly2orth

    """
    A = jax.jacfwd(poly2orth)(c, rec)
    return jnp.linalg.solve(A, c)


@jit
def orthline(off, scl, rec):
    """Orthogonal series whose graph is a straight line.

    Parameters
    ----------
    off, scl : scalars
        The specified line is given by ``off + scl*x``.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    y : ndarray
        This module's representation of the orthogonal series for
        ``off + scl*x``.

    """
    return poly2orth(jnp.array([off, scl]), rec)


@jit
def orthfromroots(roots, rec):
    """Generate an orthogonal series with given roots.

    The function returns the coefficients of the polynomial

    .. math:: p(x) = (x - r_0) * (x - r_1) * ... * (x - r_n),

    in the form of an orthogonal series with recurrence relation ``rec``, where the
    `r_n` are the roots specified in `roots`.
    If a zero has multiplicity n, then it must appear in `roots` n times.
    For instance, if 2 is a root of multiplicity three and 3 is a root of
    multiplicity 2, then `roots` looks something like [2, 2, 2, 3, 3]. The
    roots can appear in any order.

    If the returned coefficients are `c`, then

    .. math:: p(x) = c_0 + c_1 * P_1(x) + ... +  c_n * P_n(x)

    Parameters
    ----------
    roots : array_like
        Sequence containing the roots.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    out : ndarray
        1-D array of coefficients.  If all roots are real then `out` is a
        real array, if some of the roots are complex, then `out` is complex
        even if all the coefficients in the result are real (see Examples
        below).

    """
    return pu._fromroots(
        functools.partial(orthline, rec=rec),
        functools.partial(orthmul, rec=rec),
        roots,
    )


@jit
def orthadd(c1, c2, rec):
    """Add one orthogonal series to another.

    Returns the sum of two orthogonal series `c1` + `c2`.  The arguments
    are sequences of coefficients ordered from lowest order term to
    highest, i.e., [1,2,3] represents the series ``P_0 + 2*P_1 + 3*P_2``.

    Parameters
    ----------
    c1, c2 : array_like
        1-D arrays of orthogonal series coefficients ordered from low to
        high.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    out : ndarray
        Array representing the orthogonal series of their sum.

    See Also
    --------
    orthsub, orthmulx, orthmul, orthdiv, orthpow

    """
    return pu._add(c1, c2)


@jit
def orthsub(c1, c2, rec):
    """Subtract one orthogonal series from another.

    Returns the difference of two orthogonal series `c1` - `c2`.  The
    sequences of coefficients are from lowest order term to highest, i.e.,
    [1,2,3] represents the series ``P_0 + 2*P_1 + 3*P_2``.

    Parameters
    ----------
    c1, c2 : array_like
        1-D arrays of orthogonal series coefficients ordered from low to
        high.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    out : ndarray
        Of orthogonal series coefficients representing their difference.

    See Also
    --------
    orthadd, orthmulx, orthmul, orthdiv, orthpow

    """
    return pu._sub(c1, c2)


@functools.partial(jit, static_argnames="mode")
def orthmulx(c, rec, mode="full"):
    """Multiply an orthogonal series by x.

    Multiply the polynomial `c` by x, where x is the independent
    variable.

    Parameters
    ----------
    c : array_like
        1-D array of series coefficients ordered from low to
        high.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.
    mode : {"full", "same"}
        If "full", output has shape (len(c) + 1). If "same", output has shape
        (len(c)), possibly truncating high order modes.

    Returns
    -------
    out : ndarray
        Array representing the result of the multiplication.
    """
    n = jnp.arange(len(c) + 1)
    a = rec.a(n)
    b = rec.b(n)
    m = rec.m(n)

    prd = jnp.pad(c, (0, 1))
    # f = c*(m*p) where p is monic polynomial
    prd *= m  # convert c to monic form

    diagonal = a
    lower_diagonal = b[1:]
    upper_diagonal = jnp.ones_like(b[1:])

    # multiplying by transpose of monic jacobi matrix, so lower and upper are swapped
    prd = tridiagmv(diagonal, upper_diagonal, lower_diagonal, prd)

    # convert back to user scale
    prd /= m

    if mode == "same":
        prd = prd[: len(c)]

    return prd


@functools.partial(jit, static_argnames="mode")
def orthmul(c1, c2, rec, mode="full"):
    """Multiply one orthogonal series by another.

    Returns the product of two series `c1` * `c2`.  The arguments
    are sequences of coefficients, from lowest order "term" to highest,
    e.g., [1,2,3] represents the series ``P_0 + 2*P_1 + 3*P_2``.

    Parameters
    ----------
    c1, c2 : array_like
        1-D arrays of orthogonal series coefficients ordered from low to
        high.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.
    mode : {"full", "same"}
        If "full", output has shape (len(c1) + len(c2) - 1). If "same", output has shape
        max(len(c1), len(c2)), possibly truncating high order modes.

    Returns
    -------
    out : ndarray
        Of series coefficients representing their product.

    See Also
    --------
    orthadd, orthsub, orthmulx, orthdiv, orthpow

    """
    # assume c1 is longer, we iterate over c2 so want that shortest
    if len(c2) > len(c1):
        c1, c2 = c2, c1

    if len(c2) == 1:
        return c1 * c2[0]

    # convert coefficients form scaled to monic form
    # ie f * p_scaled(x) = f*m * p_monic(x)
    f = c1 * rec.m(jnp.arange(c1.size))
    g = c2 * rec.m(jnp.arange(c2.size))

    # ensure leading coefficient is 1
    fi, fn = last_nonzero(f)
    gi, gm = last_nonzero(g)
    f = f / fn
    g = g / gm

    # order of polynomials
    n = len(f) - 1
    m = len(g) - 1
    N = n + m

    # elements of jacobi matrix
    nn = jnp.arange(N)
    a = rec.a(nn)  # diagonal
    b = rec.b(nn)[1:]  # lower diagonal
    c = jnp.ones_like(b)  # upper diagonal

    r1 = jnp.pad(f, (0, m - 1))
    r2 = jnp.zeros_like(r1)
    u = r1 * g[0]

    def bodyfun(i, state):
        r1, r2, u = state
        ri = tridiagmv(a - rec.a(i - 2), c, b, r1) - rec.b(i - 2) * r2
        u += g[i - 1] * ri
        r2 = r1
        r1 = ri
        return r1, r2, u

    r1, r2, u = jax.lax.fori_loop(2, m + 1, bodyfun, (r1, r2, u))

    ri = tridiagmv(a - rec.a(gi - 1), c, b, r1) - rec.b(gi - 1) * r2
    u += g[-1] * ri
    u = jnp.append(u, 0.0)
    u = u.at[fi + gi].set(1.0)

    # scale by original leading coefficients
    u *= fn * gm
    # undo rescaling to monic form
    u /= rec.m(jnp.arange(N + 1))

    if mode == "same":
        u = u[: max(len(c1), len(c2))]

    return u


@jit
def orthdiv(c1, c2, rec):
    """Divide one orthogonal series by another.

    Returns the quotient-with-remainder of two orthogonal series
    `c1` / `c2`.  The arguments are sequences of coefficients from lowest
    order "term" to highest, e.g., [1,2,3] represents the series
    ``P_0 + 2*P_1 + 3*P_2``.

    Parameters
    ----------
    c1, c2 : array_like
        1-D arrays of orthogonal series coefficients ordered from low to
        high.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    [quo, rem] : ndarrays
        Of orthogonal series coefficients representing the quotient and
        remainder.

    See Also
    --------
    orthadd, orthsub, orthmulx, orthdiv, orthpow

    Notes
    -----
    In general, the (polynomial) division of one orthogonal series by another
    results in quotient and remainder terms that are not in the original
    polynomial basis set.  Thus, to express these results as an orthogonal
    series, it is necessary to "reproject" the results onto the original
    basis set, which may produce "unintuitive" (but correct) results.

    """
    return pu._div(functools.partial(orthmul, rec=rec), c1, c2)


@functools.partial(jit, static_argnames=("pow", "maxpower"))
def orthpow(c, pow, rec, maxpower=16):
    """Raise an orthogonal series to a power.

    Returns the orthogonal series `c` raised to the power `pow`. The
    argument `c` is a sequence of coefficients ordered from low to high.
    i.e., [1,2,3] is the series  ``P_0 + 2*P_1 + 3*P_2.``

    Parameters
    ----------
    c : array_like
        1-D array of orthogonal series coefficients ordered from low to
        high.
    pow : integer
        Power to which the series will be raised
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.
    maxpower : integer, optional
        Maximum power allowed. This is mainly to limit growth of the series
        to unmanageable size. Default is 16

    Returns
    -------
    coef : ndarray
        Orthogonal series of power.

    See Also
    --------
    orthadd, orthsub, orthmulx, orthmul, orthdiv

    """
    return pu._pow(functools.partial(orthmul, rec=rec), c, pow, maxpower)


@functools.partial(jit, static_argnames=("tensor",))
def orthval(x, c, rec, tensor=True):
    """Evaluate an orthogonal series.

    Parameters
    ----------
    x : jax.Array
        Evaluation points.
    c : jax.Array
        Coefficients of series, in ascending order
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials
    tensor : boolean, optional
        If True, the shape of the coefficient array is extended with ones
        on the right, one for each dimension of `x`. Scalars have dimension 0
        for this action. The result is that every column of coefficients in
        `c` is evaluated for every element of `x`. If False, `x` is broadcast
        over the columns of `c` for the evaluation.  This keyword is useful
        when `c` is multidimensional. The default value is True.

    Returns
    -------
    y : jax.Array
        Series evaluated at x
    """
    c = jnp.asarray(c)
    x = jnp.asarray(x)
    if tensor:
        c = c.reshape(c.shape + (1,) * x.ndim)

    c1 = c[-2] * jnp.zeros_like(x)
    c2 = c[-1] * jnp.zeros_like(x)

    def body(i, val):
        k = len(c) - i
        c1, c2 = val
        ck_plus1 = c1
        # on first step c2 is 0 so we don't care about rec.b(k+1), but we don't
        # want to try to get it to avoid out of bounds errors on tabulated coeffs.
        bk1 = rec.b(jnp.minimum(k + 1, len(c) - 1))
        ck = c[k] * rec.m(k) + (x - rec.a(k)) * c1 - bk1 * c2
        return ck, ck_plus1

    c1, c2 = jax.lax.fori_loop(1, len(c), body, (c1, c2))

    p0 = 1
    p1 = (x - rec.a(0)) * p0
    return p0 * c[0] * rec.m(0) + p1 * c1 - rec.b(1) * p0 * c2


@jit
def orthval2d(x, y, c, rec):
    r"""Evaluate a 2-D orthogonal series at points (x, y).

    This function returns the values:

    .. math:: p(x,y) = \sum_{i,j} c_{i,j} * P_i(x) * P_j(y)

    The parameters `x` and `y` are converted to arrays only if they are
    tuples or a lists, otherwise they are treated as a scalars and they
    must have the same shape after conversion. In either case, either `x`
    and `y` or their elements must support multiplication and addition both
    with themselves and with the elements of `c`.

    If `c` is a 1-D array a one is implicitly appended to its shape to make
    it 2-D. The shape of the result will be c.shape[2:] + x.shape.

    Parameters
    ----------
    x, y : array_like, compatible objects
        The two dimensional series is evaluated at the points `(x, y)`,
        where `x` and `y` must have the same shape. If `x` or `y` is a list
        or tuple, it is first converted to an ndarray, otherwise it is left
        unchanged and if it isn't an ndarray it is treated as a scalar.
    c : array_like
        Array of coefficients ordered so that the coefficient of the term
        of multi-degree i,j is contained in ``c[i,j]``. If `c` has
        dimension greater than two the remaining indices enumerate multiple
        sets of coefficients.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials

    Returns
    -------
    values : ndarray, compatible object
        The values of the two dimensional polynomial at points formed with
        pairs of corresponding values from `x` and `y`.

    See Also
    --------
    orthval, orthgrid2d, orthval3d, orthgrid3d

    """
    return pu._valnd(functools.partial(orthval, rec=rec), c, x, y)


@jit
def orthgrid2d(x, y, c, rec):
    r"""Evaluate a 2-D orthogonal series on the Cartesian product of x and y.

    This function returns the values:

    .. math:: p(a,b) = \sum_{i,j} c_{i,j} * P_i(a) * P_j(b)

    where the points `(a, b)` consist of all pairs formed by taking
    `a` from `x` and `b` from `y`. The resulting points form a grid with
    `x` in the first dimension and `y` in the second.

    The parameters `x` and `y` are converted to arrays only if they are
    tuples or a lists, otherwise they are treated as a scalars. In either
    case, either `x` and `y` or their elements must support multiplication
    and addition both with themselves and with the elements of `c`.

    If `c` has fewer than two dimensions, ones are implicitly appended to
    its shape to make it 2-D. The shape of the result will be c.shape[2:] +
    x.shape + y.shape.

    Parameters
    ----------
    x, y : array_like, compatible objects
        The two dimensional series is evaluated at the points in the
        Cartesian product of `x` and `y`.  If `x` or `y` is a list or
        tuple, it is first converted to an ndarray, otherwise it is left
        unchanged and, if it isn't an ndarray, it is treated as a scalar.
    c : array_like
        Array of coefficients ordered so that the coefficient of the term of
        multi-degree i,j is contained in `c[i,j]`. If `c` has dimension
        greater than two the remaining indices enumerate multiple sets of
        coefficients.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials

    Returns
    -------
    values : ndarray, compatible object
        The values of the two dimensional Chebyshev series at points in the
        Cartesian product of `x` and `y`.

    See Also
    --------
    orthval, orthval2d, orthval3d, orthgrid3d

    """
    return pu._gridnd(functools.partial(orthval, rec=rec), c, x, y)


@jit
def orthval3d(x, y, z, c, rec):
    r"""Evaluate a 3-D orthogonal series at points (x, y, z).

    This function returns the values:

    .. math:: p(x,y,z) = \sum_{i,j,k} c_{i,j,k} * P_i(x) * P_j(y) * P_k(z)

    The parameters `x`, `y`, and `z` are converted to arrays only if
    they are tuples or a lists, otherwise they are treated as a scalars and
    they must have the same shape after conversion. In either case, either
    `x`, `y`, and `z` or their elements must support multiplication and
    addition both with themselves and with the elements of `c`.

    If `c` has fewer than 3 dimensions, ones are implicitly appended to its
    shape to make it 3-D. The shape of the result will be c.shape[3:] +
    x.shape.

    Parameters
    ----------
    x, y, z : array_like, compatible object
        The three dimensional series is evaluated at the points
        `(x, y, z)`, where `x`, `y`, and `z` must have the same shape.  If
        any of `x`, `y`, or `z` is a list or tuple, it is first converted
        to an ndarray, otherwise it is left unchanged and if it isn't an
        ndarray it is  treated as a scalar.
    c : array_like
        Array of coefficients ordered so that the coefficient of the term of
        multi-degree i,j,k is contained in ``c[i,j,k]``. If `c` has dimension
        greater than 3 the remaining indices enumerate multiple sets of
        coefficients.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials

    Returns
    -------
    values : ndarray, compatible object
        The values of the multidimensional polynomial on points formed with
        triples of corresponding values from `x`, `y`, and `z`.

    See Also
    --------
    orthval, orthval2d, orthgrid2d, orthgrid3d

    """
    return pu._valnd(functools.partial(orthval, rec=rec), c, x, y, z)


@jit
def orthgrid3d(x, y, z, c, rec):
    r"""Evaluate a 3-D orthogonal series on the Cartesian product of x, y, and z.

    This function returns the values:

    .. math:: p(a,b,c) = \sum_{i,j,k} c_{i,j,k} * P_i(a) * P_j(b) * P_k(c)

    where the points `(a, b, c)` consist of all triples formed by taking
    `a` from `x`, `b` from `y`, and `c` from `z`. The resulting points form
    a grid with `x` in the first dimension, `y` in the second, and `z` in
    the third.

    The parameters `x`, `y`, and `z` are converted to arrays only if they
    are tuples or a lists, otherwise they are treated as a scalars. In
    either case, either `x`, `y`, and `z` or their elements must support
    multiplication and addition both with themselves and with the elements
    of `c`.

    If `c` has fewer than three dimensions, ones are implicitly appended to
    its shape to make it 3-D. The shape of the result will be c.shape[3:] +
    x.shape + y.shape + z.shape.

    Parameters
    ----------
    x, y, z : array_like, compatible objects
        The three dimensional series is evaluated at the points in the
        Cartesian product of `x`, `y`, and `z`.  If `x`,`y`, or `z` is a
        list or tuple, it is first converted to an ndarray, otherwise it is
        left unchanged and, if it isn't an ndarray, it is treated as a
        scalar.
    c : array_like
        Array of coefficients ordered so that the coefficients for terms of
        degree i,j are contained in ``c[i,j]``. If `c` has dimension
        greater than two the remaining indices enumerate multiple sets of
        coefficients.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials

    Returns
    -------
    values : ndarray, compatible object
        The values of the two dimensional polynomial at points in the Cartesian
        product of `x` and `y`.

    See Also
    --------
    orthval, orthval2d, orthgrid2d, orthval3d

    """
    return pu._gridnd(functools.partial(orthval, rec=rec), c, x, y, z)


@functools.partial(jit, static_argnames=("deg",))
def orthvander(x, deg, rec):
    """Pseudo-Vandermonde matrix of given degree.

    Returns the pseudo-Vandermonde matrix of degree `deg` and sample points
    `x`. The pseudo-Vandermonde matrix is defined by

    .. math:: V[..., i] = p_i(x),

    where `0 <= i <= deg`. The leading indices of `V` index the elements of
    `x` and the last index is the degree of the Chebyshev polynomial.

    If `c` is a 1-D array of coefficients of length `n + 1` and `V` is the
    matrix ``V = orthvander(x, n, rec)``, then ``np.dot(V, c)`` and
    ``orthval(x, c, rec)`` are the same up to roundoff.  This equivalence is
    useful both for least squares fitting and for the evaluation of a large
    number of Chebyshev series of the same degree and sample points.

    Parameters
    ----------
    x : jax.Array
        Array of points. The dtype is converted to float64 or complex128
        depending on whether any of the elements are complex. If `x` is
        scalar it is converted to a 1-D array.
    deg : int
        Degree of the resulting matrix.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    vander : jax.Array
        The pseudo Vandermonde matrix. The shape of the returned matrix is
        ``x.shape + (deg + 1,)``, where The last index is the degree of the
        corresponding polynomial.  The dtype will be the same as
        the converted `x`.

    """
    if deg < 0:
        raise ValueError("deg must be non-negative")

    x = jnp.array(x, ndmin=1)
    dims = (deg + 1,) + x.shape
    dtyp = jnp.promote_types(x.dtype, jnp.array(0.0).dtype)
    x = x.astype(dtyp)
    v = jnp.empty(dims, dtype=dtyp)
    # easiest to evaluate in monic form then scale at the end
    v = v.at[0].set(jnp.ones_like(x))
    if deg > 0:
        v = v.at[1].set(x - rec.a(0))

        def body(i, v):
            p0 = v[i - 2]
            p1 = v[i - 1]
            pn = (x - rec.a(i - 1)) * p1 - rec.b(i - 1) * p0
            v = v.at[i].set(pn)
            return v

        v = jax.lax.fori_loop(2, deg + 1, body, v)

    return jnp.moveaxis(v, 0, -1) * rec.m(jnp.arange(deg + 1))


@functools.partial(jit, static_argnames=("deg",))
def orthvander2d(x, y, deg, rec):
    r"""Pseudo-Vandermonde matrix of given degrees.

    Returns the pseudo-Vandermonde matrix of degrees `deg` and sample
    points `(x, y)`. The pseudo-Vandermonde matrix is defined by

    .. math:: V[..., (deg[1] + 1)*i + j] = P_i(x) * P_j(y),

    where `0 <= i <= deg[0]` and `0 <= j <= deg[1]`. The leading indices of
    `V` index the points `(x, y)` and the last index encodes the degrees of
    the orthogonal polynomials.

    If ``V = orthvander2d(x, y, [xdeg, ydeg])``, then the columns of `V`
    correspond to the elements of a 2-D coefficient array `c` of shape
    (xdeg + 1, ydeg + 1) in the order

    .. math:: c_{00}, c_{01}, c_{02} ... , c_{10}, c_{11}, c_{12} ...

    and ``np.dot(V, c.flat)`` and ``orthval2d(x, y, c)`` will be the same
    up to roundoff. This equivalence is useful both for least squares
    fitting and for the evaluation of a large number of 2-D orthogonal
    series of the same degrees and sample points.

    Parameters
    ----------
    x, y : array_like
        Arrays of point coordinates, all of the same shape. The dtypes
        will be converted to either float64 or complex128 depending on
        whether any of the elements are complex. Scalars are converted to
        1-D arrays.
    deg : tuple of ints
        Tuple of maximum degrees of the form (x_deg, y_deg).
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    vander2d : ndarray
        The shape of the returned matrix is ``x.shape + (order,)``, where
        :math:`order = (deg[0]+1)*(deg[1]+1)`.  The dtype will be the same
        as the converted `x` and `y`.

    See Also
    --------
    orthvander, orthvander3d, orthval2d, orthval3d

    """
    vander = functools.partial(orthvander, rec=rec)
    return pu._vander_nd_flat((vander, vander), (x, y), deg)


@functools.partial(jit, static_argnames=("deg",))
def orthvander3d(x, y, z, deg, rec):
    r"""Pseudo-Vandermonde matrix of given degrees.

    Returns the pseudo-Vandermonde matrix of degrees `deg` and sample
    points `(x, y, z)`. If `l, m, n` are the given degrees in `x, y, z`,
    then The pseudo-Vandermonde matrix is defined by

    .. math:: V[..., (m+1)(n+1)i + (n+1)j + k] = P_i(x)*P_j(y)*P_k(z),

    where `0 <= i <= l`, `0 <= j <= m`, and `0 <= j <= n`.  The leading
    indices of `V` index the points `(x, y, z)` and the last index encodes
    the degrees of the orthogonal polynomials.

    If ``V = orthvander3d(x, y, z, [xdeg, ydeg, zdeg])``, then the columns
    of `V` correspond to the elements of a 3-D coefficient array `c` of
    shape (xdeg + 1, ydeg + 1, zdeg + 1) in the order

    .. math:: c_{000}, c_{001}, c_{002},... , c_{010}, c_{011}, c_{012},...

    and  ``np.dot(V, c.flat)`` and ``lagval3d(x, y, z, c)`` will be the
    same up to roundoff. This equivalence is useful both for least squares
    fitting and for the evaluation of a large number of 3-D orthogonal
    series of the same degrees and sample points.

    Parameters
    ----------
    x, y, z : array_like
        Arrays of point coordinates, all of the same shape. The dtypes will
        be converted to either float64 or complex128 depending on whether
        any of the elements are complex. Scalars are converted to 1-D
        arrays.
    deg : tuple of ints
        Tuple of maximum degrees of the form (x_deg, y_deg, z_deg).
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    vander3d : ndarray
        The shape of the returned matrix is ``x.shape + (order,)``, where
        :math:`order = (deg[0]+1)*(deg[1]+1)*(deg[2]+1)`.  The dtype will
        be the same as the converted `x`, `y`, and `z`.

    See Also
    --------
    orthvander, orthvander3d, orthval2d, orthval3d

    """
    vander = functools.partial(orthvander, rec=rec)
    return pu._vander_nd_flat((vander, vander, vander), (x, y, z), deg)


@functools.partial(jit, static_argnames=("deg", "full"))
def orthfit(x, y, deg, rec, rcond=None, full=False, w=None):
    r"""Least squares fit of orthogonal series to data.

    Return the coefficients of an orthogonal series of degree `deg` that is the
    least squares fit to the data values `y` given at points `x`. If `y` is
    1-D the returned coefficients will also be 1-D. If `y` is 2-D multiple
    fits are done, one for each column of `y`, and the resulting
    coefficients are stored in the corresponding columns of a 2-D return.
    The fitted polynomial(s) are in the form

    .. math::  p(x) = c_0 + c_1 * P_1(x) + ... + c_n * P_n(x),

    where ``n`` is `deg`.

    Parameters
    ----------
    x : array_like, shape (M,)
        x-coordinates of the M sample points ``(x[i], y[i])``.
    y : array_like, shape (M,) or (M, K)
        y-coordinates of the sample points. Several data sets of sample
        points sharing the same x-coordinates can be fitted at once by
        passing in a 2D-array that contains one dataset per column.
    deg : int or 1-D array_like
        Degree(s) of the fitting polynomials. If `deg` is a single integer
        all terms up to and including the `deg`'th term are included in the
        fit. A tuple of integers specifying the degrees of the terms to include
        may be used instead.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.
    rcond : float, optional
        Relative condition number of the fit. Singular values smaller than
        this relative to the largest singular value will be ignored. The
        default value is len(x)*eps, where eps is the relative precision of
        the float type, about 2e-16 in most cases.
    full : bool, optional
        Switch determining nature of return value. When it is False (the
        default) just the coefficients are returned, when True diagnostic
        information from the singular value decomposition is also returned.
    w : array_like, shape (`M`,), optional
        Weights. If not None, the weight ``w[i]`` applies to the unsquared
        residual ``y[i] - y_hat[i]`` at ``x[i]``. Ideally the weights are
        chosen so that the errors of the products ``w[i]*y[i]`` all have the
        same variance.  When using inverse-variance weighting, use
        ``w[i] = 1/sigma(y[i])``.  The default value is None.

    Returns
    -------
    coef : ndarray, shape (M,) or (M, K)
        Orthogonal coefficients ordered from low to high. If `y` was 2-D,
        the coefficients for the data in column *k*  of `y` are in column
        *k*.

    [residuals, rank, singular_values, rcond] : list
        These values are only returned if ``full == True``

        - residuals -- sum of squared residuals of the least squares fit
        - rank -- the numerical rank of the scaled Vandermonde matrix
        - singular_values -- singular values of the scaled Vandermonde matrix
        - rcond -- value of `rcond`.

        For more details, see `jax.numpy.linalg.lstsq`.

    See Also
    --------
    orthax.polynomial.polyfit
    orthax.laguerre.lagfit
    orthax.legendre.legfit
    orthax.chebyshev.chebfit
    orthax.hermite.hermfit
    orthax.hermite_e.hermefit
    orthval : Evaluates an orthogonal series.
    orthvander : pseudo Vandermonde matrix of orthogonal series.
    jax.numpy.linalg.lstsq : Computes a least-squares fit from the matrix.

    Notes
    -----
    The solution is the coefficients of the orthogonal series ``p`` that
    minimizes the sum of the weighted squared errors

    .. math:: E = \sum_j w_j^2 * |y_j - p(x_j)|^2,

    where the :math:`w_j` are the weights. This problem is solved by
    setting up as the (typically) overdetermined matrix equation

    .. math:: V(x) * c = w * y,

    where ``V`` is the weighted pseudo Vandermonde matrix of `x`, ``c`` are the
    coefficients to be solved for, `w` are the weights, and `y` are the
    observed values.  This equation is then solved using the singular value
    decomposition of ``V``.

    If some of the singular values of `V` are so small that they are
    neglected, then a `RankWarning` will be issued. This means that the
    coefficient values may be poorly determined. Using a lower order fit
    will usually get rid of the warning.  The `rcond` parameter can also be
    set to a value smaller than its default, but the resulting fit may be
    spurious and have large contributions from roundoff error.

    Fits using orthogonal series are probably most useful when the data can
    be approximated by ``sqrt(w(x)) * p(x)``, where ``w(x)`` is the series
    weight function. In that case the weight ``sqrt(w(x[i]))`` should be used
    together with data values ``y[i]/sqrt(w(x[i]))``. The weight function is
    available as `rec.weight`.

    """
    vander = functools.partial(orthvander, rec=rec)
    return pu._fit(vander, x, y, deg, rcond, full, w)


@jit
def orthweight(x, rec):
    r"""Weight function of orthogonal polynomials.

    Parameters
    ----------
    x : array_like
       Values at which the weight function will be computed.

    Returns
    -------
    w : ndarray
       The weight function at `x`.

    """
    return rec.weight(x)


@jit
def orthnorm(n, rec):
    r"""Norm of nth orthogonal polynomial.

    The norm :math:`\gamma_n` is defined such that

    :math:`\int_{a}^{b} P_n^2(x) w(x) dx = \gamma_n^2`

    Parameters
    ----------
    n : int
       Order of orthogonal polynomial.

    Returns
    -------
    gamma_n : float
       Norm of the nth orthogonal polynomial.

    """
    return rec.g(n) * jnp.abs(rec.m(n))


@functools.partial(jit, static_argnames=("monic",))
def polyval(x, n, rec, monic=False):
    """Evaluate the nth order polynomial from the family given by rec

    Parameters
    ----------
    x : jax.Array
        Evaluation points.
    n : int
        Order of polynomial
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials.

    Returns
    -------
    y : jax.Array
        Polynomial evaluated at x
    """
    x = jnp.asarray(x)

    p0 = jnp.zeros_like(x)
    p1 = jnp.ones_like(x)
    pn = p1

    def body(i, state):
        p0, p1, pn = state
        pn = (x - rec.a(i)) * p1 - rec.b(i) * p0
        p0 = p1
        p1 = pn
        return p0, p1, pn

    npos = lambda: jax.lax.fori_loop(0, n, body, (p0, p1, pn))[-1]
    nneg = lambda: jnp.zeros_like(x)
    out = jax.lax.cond(n >= 0, npos, nneg)
    if not monic:
        out *= rec.m(n)
    return out


def differentiation_matrix(rec, n):
    """Differentiation matrix for series up to degree n

    Parameters
    ----------
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials
    n : int
        Maximum order of polynomial

    Returns
    -------
    D : jax.Array, shape(n+1,n+1)
        Differentiation matrix, such that f' = D@f
    """
    a = rec.a
    b = rec.b
    m = rec.m
    n += 1
    nn = jnp.arange(n)
    Q = jnp.diag(nn.astype(float))

    def iloop(i, Q):
        def jloop(j, Q):
            im1 = jnp.maximum(i - 1, 0)
            jm1 = jnp.maximum(j - 1, 0)
            Q = Q.at[i, j].add(((j > 0) * a(jm1) - (i > 0) * a(im1)) * Q[i - 1, j])
            Q = Q.at[i, j].add((j > 1) * Q[i - 1, j - 1])
            Q = Q.at[i, j].add(b(j) * Q[i - 1, j + 1])
            Q = Q.at[i, j].add(-1 * (i > 1) * b(im1) * Q[i - 2, j])
            return Q

        return jax.lax.fori_loop(0, i, jloop, Q)

    Q = jax.lax.fori_loop(0, n, iloop, Q)
    D = jnp.pad(Q[1:, 1:], ((1, 0), (0, 1))).T
    D = 1 / m(nn)[:, None] * D * m(nn)
    return D


def integration_matrix(rec, n):
    """Integration matrix for series up to degree n

    Parameters
    ----------
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials
    n : int
        Maximum order of polynomial

    Returns
    -------
    S : jax.Array, shape(n,n)
        Integration matrix, such that f = S@f'
    """
    D = differentiation_matrix(rec, n)
    return jnp.linalg.pinv(D)


def _pad_along_axis(array: jax.Array, pad: tuple = (0, 0), axis: int = 0):
    """Pad with zeros or truncate a given dimension."""
    array = jnp.moveaxis(array, axis, 0)

    if pad[0] < 0:
        array = array[abs(pad[0]) :]
        pad = (0, pad[1])
    if pad[1] < 0:
        array = array[: -abs(pad[1])]
        pad = (pad[0], 0)

    npad = [(0, 0)] * array.ndim
    npad[0] = pad

    array = jnp.pad(array, pad_width=npad, mode="constant", constant_values=0)
    return jnp.moveaxis(array, 0, axis)


@functools.partial(jit, static_argnames=("m", "axis"))
def orthder(c, rec, m=1, scl=1, axis=0):
    """Differentiate an orthogonal series.

    Returns the orthogonal series coefficients `c` differentiated `m` times
    along `axis`.  At each iteration the result is multiplied by `scl` (the
    scaling factor is for use in a linear change of variable). The argument
    `c` is an array of coefficients from low to high degree along each
    axis, e.g., [1,2,3] represents the series ``1*P_0 + 2*P_1 + 3*P_2``
    while [[1,2],[1,2]] represents ``1*P_0(x)*P_0(y) + 1*P_1(x)*P_0(y) +
    2*P_0(x)*P_1(y) + 2*P_1(x)*P_1(y)`` if axis=0 is ``x`` and axis=1 is
    ``y``.

    Parameters
    ----------
    c : array_like
        Array of orthogonal series coefficients. If c is multidimensional the
        different axis correspond to different variables with the degree in
        each axis given by the corresponding index.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials
    m : int, optional
        Number of derivatives taken, must be non-negative. (Default: 1)
    scl : scalar, optional
        Each differentiation is multiplied by `scl`.  The end result is
        multiplication by ``scl**m``.  This is for use in a linear change of
        variable. (Default: 1)
    axis : int, optional
        Axis over which the derivative is taken. (Default: 0).

    Returns
    -------
    der : ndarray
        Orthogonal series of the derivative.

    See Also
    --------
    orthint

    """
    if m < 0:
        raise ValueError("The order of derivation must be non-negative")

    c = pu.as_series(c)

    if m == 0:
        return c

    c = jnp.moveaxis(c, axis, 0)
    n = len(c)
    if m >= n:
        c = jnp.zeros_like(c[:1])
    else:
        D = differentiation_matrix(rec, len(c) - 1)
        # TODO: figure out how to get rid of this python loop
        for i in range(m):
            c = D @ c * scl

    c = c[:-m]
    c = jnp.moveaxis(c, 0, axis)
    return c


@functools.partial(jit, static_argnames=("m", "axis"))
def orthint(c, rec, m=1, k=[], lbnd=0, scl=1, axis=0):
    """Integrate an orthogonal series.

    Returns the orthogonal series coefficients `c` integrated `m` times from
    `lbnd` along `axis`. At each iteration the resulting series is
    **multiplied** by `scl` and an integration constant, `k`, is added.
    The scaling factor is for use in a linear change of variable.  ("Buyer
    beware": note that, depending on what one is doing, one may want `scl`
    to be the reciprocal of what one might expect; for more information,
    see the Notes section below.)  The argument `c` is an array of
    coefficients from low to high degree along each axis, e.g., [1,2,3]
    represents the series ``P_0 + 2*P_1 + 3*P_2`` while [[1,2],[1,2]]
    represents ``1*P_0(x)*P_0(y) + 1*P_1(x)*P_0(y) + 2*P_0(x)*P_1(y) +
    2*P_1(x)*P_1(y)`` if axis=0 is ``x`` and axis=1 is ``y``.

    Parameters
    ----------
    c : array_like
        Array of orthogonal series coefficients. If c is multidimensional the
        different axis correspond to different variables with the degree in
        each axis given by the corresponding index.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials
    m : int, optional
        Order of integration, must be positive. (Default: 1)
    k : {[], list, scalar}, optional
        Integration constant(s).  The value of the first integral at
        ``lbnd`` is the first value in the list, the value of the second
        integral at ``lbnd`` is the second value, etc.  If ``k == []`` (the
        default), all constants are set to zero.  If ``m == 1``, a single
        scalar can be given instead of a list.
    lbnd : scalar, optional
        The lower bound of the integral. (Default: 0)
    scl : scalar, optional
        Following each integration the result is *multiplied* by `scl`
        before the integration constant is added. (Default: 1)
    axis : int, optional
        Axis over which the integral is taken. (Default: 0).

    Returns
    -------
    S : ndarray
        Orthogonal series coefficient array of the integral.

    Raises
    ------
    ValueError
        If ``m < 0``, ``len(k) > m``, ``np.ndim(lbnd) != 0``, or
        ``np.ndim(scl) != 0``.

    See Also
    --------
    orthder

    Notes
    -----
    Note that the result of each integration is *multiplied* by `scl`.
    Why is this important to note?  Say one is making a linear change of
    variable :math:`u = ax + b` in an integral relative to `x`.  Then
    :math:`dx = du/a`, so one will need to set `scl` equal to
    :math:`1/a` - perhaps not what one would have first thought.

    """
    c = pu.as_series(c)
    lbnd, scl = map(jnp.asarray, (lbnd, scl))

    if not jnp.iterable(k):
        k = [k]
    if len(k) > m:
        raise ValueError("Too many integration constants")
    if jnp.ndim(lbnd) != 0:
        raise ValueError("lbnd must be a scalar.")
    if jnp.ndim(scl) != 0:
        raise ValueError("scl must be a scalar.")

    if m == 0:
        return c

    c = jnp.moveaxis(c, axis, 0)
    k = jnp.array(list(k) + [0] * (m - len(k)), ndmin=1)

    I = integration_matrix(rec, len(c) + m - 1)
    c = _pad_along_axis(c, (0, m), axis=0)
    # TODO: figure out how to get rid of this python loop
    for i in range(m):
        c *= scl
        c = I @ c
        c = c.at[0].add(k[i] - orthval(lbnd, c, rec))
    c = jnp.moveaxis(c, 0, axis)
    return c


def jacobi_matrix(rec, n):
    """Return the Jacobi matrix for a given set of orthogonal polynomials.


    Parameters
    ----------
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials
    n : int
        Degree of Jacobi matrix

    Returns
    -------
    J : jax.Array, shape(n,n)
        Jacobi matrix.
    """
    nn = jnp.arange(n)
    a = rec.a(nn)
    b = rec.b(nn)
    return (
        jnp.diag(a) + jnp.diag(jnp.sqrt(b[1:]), k=1) + jnp.diag(jnp.sqrt(b[1:]), k=-1)
    )


def orthgauss(deg, rec, x0=None, x1=None):
    """Compute Gaussian quadrature nodes and weights for given orthogonal polynomials.

    Can optionally compute Gauss-Radau or Gauss-Lobatto points if x0 and/or x1 are
    given.

    Parameters
    ----------
    deg : int
        Number of sample points and weights. It must be >= 1.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials
    x0, x1 : float, optional
        Optional points to include in the nodes, for Gauss-Radau (if x0 is given) or
        Gauss-Lobatto (if x0 and x1 given). Generally taken to be the start and/or end
        points of the interval of integration.

    Returns
    -------
    x, w : jax.Array, shape(n,)
        Nodes and weights for n point quadrature rule.
    """
    n = deg - (x0 is not None) - (x1 is not None)
    J = jacobi_matrix(rec, n)
    if x0 is not None:
        x0 = jnp.asarray(x0).astype(float)
        J = jnp.pad(J, ((0, 1), (0, 1)))
        J = J.at[n, n - 1].set(jnp.sqrt(rec.b(n)))
        J = J.at[n - 1, n].set(jnp.sqrt(rec.b(n)))
        if x1 is None:
            # gauss radau
            astar = x0 - rec.b(n) * polyval(x0, n - 1, rec, True) / polyval(
                x0, n, rec, True
            )
            J = J.at[n, n].set(astar)
        else:
            # gauss lobatto
            x1 = jnp.asarray(x1).astype(float)
            J = J.at[n, n].set(rec.a(n))
            mat = jnp.array(
                [
                    [polyval(x0, n + 1, rec, True), polyval(x0, n, rec, True)],
                    [polyval(x1, n + 1, rec, True), polyval(x1, n, rec, True)],
                ]
            )
            rhs = jnp.array(
                [x0 * polyval(x0, n + 1, rec, True), x1 * polyval(x1, n + 1, rec, True)]
            )
            astar, bstar = jnp.linalg.solve(mat, rhs)
            J = jnp.pad(J, ((0, 1), (0, 1)))
            J = J.at[n + 1, n + 1].set(astar)
            J = J.at[n, n + 1].set(jnp.sqrt(bstar))
            J = J.at[n + 1, n].set(jnp.sqrt(bstar))
    else:
        assert x1 is None
    x, v = jnp.linalg.eigh(J)
    w = rec.b(0) * v[0] ** 2
    return x, w


@jit
def orthcompanion(c, rec):
    """Return the companion matrix of c.

    Parameters
    ----------
    c : array_like
        1-D array of orthogonal series coefficients ordered from low to high
        degree.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials

    Returns
    -------
    mat : jax.Array
        Companion matrix of dimensions (deg, deg).

    """
    c = pu.as_series(c)

    if len(c) < 2:
        raise ValueError("Series must have maximum degree of at least 1.")

    nn = jnp.arange(len(c))
    # convert to monic form
    c = c * rec.m(nn)
    c = c / c[-1]  # divide out leading coeff
    # based on
    # "A companion matrix analogue for orthogonal polynomials", S. Barnett, 1975

    a = rec.a(nn)
    b = rec.b(nn)
    alpha = jnp.ones_like(a)
    beta = -a
    gamma = b
    A = (
        jnp.diag(-beta[:-1] / alpha[:-1])
        + jnp.diag(1 / alpha[:-2], k=1)
        + jnp.diag(gamma[1:-1] / alpha[1:-1], k=-1)
    )
    return A.at[-1, :].add(-c[:-1] / alpha[-1])


@jit
def orthroots(c, rec):
    r"""Compute the roots of an orthogonal series.

    Return the roots (a.k.a. "zeros") of the polynomial

    .. math:: p(x) = \sum_i c[i] * P_i(x).

    Parameters
    ----------
    c : 1-D array_like
        1-D array of coefficients.
    rec : AbstractRecurrenceRelation
        Recurrence relation for the family of orthogonal polynomials

    Returns
    -------
    out : ndarray
        Array of the roots of the series.

    Notes
    -----
    The root estimates are obtained as the eigenvalues of the companion
    matrix, Roots far from the origin of the complex plane may have large
    errors due to the numerical instability of the series for such values.
    Roots with multiplicity greater than 1 will also show larger errors as
    the value of the series near such points is relatively insensitive to
    errors in the roots. Isolated roots near the origin can be improved by
    a few iterations of Newton's method.

    """
    c = pu.as_series(c)
    if len(c) <= 1:
        return jnp.array([], dtype=c.dtype)

    # rotated companion matrix reduces error
    m = orthcompanion(c, rec)[::-1, ::-1]
    r = jnp.linalg.eigvals(m)
    r = jnp.sort(r)
    # TODO: add newton iterations to improve roots
    return r



r"""
===============================
Three Term Recurrence Relations
===============================

Every family of orthogonal polynomials can be shown to satisfy a three term recurrence
relation of the form

.. math::

    \begin{align}
    p_{-1}(x) &= 0 \\
    p_0(x) &= 1 \\
    p_{i+1}(x) &= (x-a_i) p_i(x) - b_i p_{i-1}(x) \\
    \end{align}

Knowing the coefficients of the recurrence relation :math:`a_i, b_i` along with the
normalization constants :math:`g_i` such that

.. math::

    \int_D p_i(x) p_j(x) w(x) dx = g^2_i \delta_{i,j}


provides all the needed information for performing arithmetic, calculus, and other
manipulations on orthogonal polynomial series.


This module provides a number of classes for storing information about these recurrence
relationships that can be used with any of the functions in the base ``orthax``
namespace that expect an ``AbstractRecurrenceRelation``. These include classes
representing many of the "classical" orthogonal polynomial families, as well as the
function ``generate_recurrence`` for generating recurrence coefficients for orthogonal
polynomials with arbitrary user defined weight functions and domains.


Functions
---------
.. autosummary::
   :toctree: generated/

   generate_recurrence

Base Classes
------------
.. autosummary::
   :toctree: generated/

   AbstractRecurrenceRelation
   TabulatedRecurrenceRelation

Classical Recurrence Relations
------------------------------
.. autosummary::
   :toctree: generated/

   Legendre
   ShiftedLegendre
   ChebyshevT
   ChebyshevU
   ChebyshevV
   ChebyshevW
   Gegenbauer
   Jacobi
   Laguerre
   GeneralizedLaguerre
   Hermite
   HermiteE

"""

import abc

import equinox as eqx
import jax
import jax.numpy as jnp
from jax.scipy.special import gammaln


def _asarray(k, kmax=None):
    k = jnp.asarray(k)
    k = eqx.error_if(
        k,
        (k < 0).any(),
        "Negative indices not allowed for recurrence coefficients.",
    )
    if kmax is not None:
        k = eqx.error_if(
            k,
            (k > kmax).any(),
            "Requested recurrence coefficient outside of tabulated range.",
        )

    return k


class AbstractRecurrenceRelation(eqx.Module, abc.ABC):
    """Base class for three term recurrence relations.

    Subclasses should declare attributes `_weight` and `_domain` and implement methods
    `a`, `b`, `g`, `m`
    """

    _weight: callable = eqx.field(static=True)
    _domain: tuple[float, float]

    @property
    def weight(self):
        """callable: Weight function defining inner product."""
        return self._weight

    @property
    def domain(self):
        """tuple: Lower and upper bounds for inner product defining orthogonality."""
        return self._domain

    @abc.abstractmethod
    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        pass

    @abc.abstractmethod
    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        pass

    @abc.abstractmethod
    def g(self, k):
        """Weighted norm of the kth monic orthogonal polynomial."""
        pass

    @abc.abstractmethod
    def m(self, k):
        """Coefficient of x**k in the kth polynomial in the desired normalization."""
        pass


class TabulatedRecurrenceRelation(AbstractRecurrenceRelation):
    """Recurrence relation from tabulated values.

    Parameters
    ----------
    weight : callable
        Weight function.
    domain : tuple
        Lower and upper bounds for inner product defining orthogonality.
    a, b : jax.Array
        Coefficients of the monic three term recurrence relation.
    g : jax.Array
        ``g[k]`` is the weighted norm of the kth monic orthogonal polynomial.
    m : jax.Array
        ``m[k]`` is the coefficient of x**k in the kth orthogonal polynomial in the
        desired normalization. Default is 1 (monic form). For normalized form, set
        m = 1/g

    """

    _ak: jax.Array
    _bk: jax.Array
    _gk: jax.Array
    _mk: jax.Array

    def __init__(self, weight, domain, ak, bk, gk, mk=None):
        if mk is None:
            mk = jnp.ones_like(ak)
        self._ak = ak
        self._bk = bk
        self._gk = gk
        self._mk = mk
        self._weight = weight
        self._domain = domain

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k, kmax=len(self._ak) - 1)
        return self._ak[k]

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k, kmax=len(self._bk) - 1)
        return self._bk[k]

    def g(self, k):
        """Weighted norm of the kth monic orthogonal polynomial."""
        k = _asarray(k, kmax=len(self._gk) - 1)
        return self._gk[k]

    def m(self, k):
        """Coefficient of x**k in the kth polynomial in the desired normalization."""
        k = _asarray(k, kmax=len(self._mk) - 1)
        return self._mk[k]


class ClassicalRecurrenceRelation(AbstractRecurrenceRelation, abc.ABC):
    """Base class for recurrence relations for "classical" orthogonal polynomials.

    Parameters
    ----------
    weight : callable
        Weight function.
    domain : tuple
        Lower and upper bounds for inner product defining orthogonality.
    scale : {"standard", "monic", "normalized"}
        Most classical orthogonal polynomials have ad-hoc normalizations (ie,
        the common definitions in textbooks are neither monic nor unit norm). This
        is encompassed in the "standard" scale, which should match that in texts such as
        Abramowitz & Stegun. Alternatively, they can be scaled to be monic or unit norm.
    """

    _scale: str = eqx.field(static=True)

    def __init__(self, weight, domain, scale="standard"):
        assert scale in {"standard", "monic", "normalized"}

        self._weight = weight
        self._domain = domain
        self._scale = scale

    @abc.abstractmethod
    def _std_norm(self, k):
        # norm of the kth polynomial in "standard" scaling (ie, AS, wikipedia, etc)
        pass

    @abc.abstractmethod
    def _std_scale(self, k):
        # coefficient of x**k in "standard" scaling (ie, AS, wikipedia, etc)
        pass

    def g(self, k):
        """Weighted norm of the kth monic orthogonal polynomial."""
        k = _asarray(k)
        return self._std_norm(k) / jnp.abs(self._std_scale(k))

    def m(self, k):
        """Coefficient of x**k in the kth polynomial in the desired normalization."""
        # scaling factor. polynomials are evaluated in monic form then multiplied
        # by this scale factor
        # ie, evaluate monic polynomials then multiply by m to get scaled version
        k = _asarray(k)
        if self._scale == "monic":
            out = jnp.ones_like(k)
        elif self._scale == "standard":
            out = self._std_scale(k)
        else:  # normalized
            out = self._std_scale(k) / self._std_norm(k)
        return out


# General notes:
# ak, bk from Gautschi, Orthogonal Polynomials: Computation and Approximation,
# Table 1.1, ak=alpha_k, bk = beta_k
# std_scale, std_norm from NIST Handbook of Mathematical Functions,
# Table 18.3.1, std_scale = k_n, std_norm = sqrt(h_n)


class Legendre(ClassicalRecurrenceRelation):
    """Recurrence relation for Legendre Polynomials :math:`P_n(x)`

    Legendre polynomials are orthogonal on the interval (-1, 1)
    with the weight function :math:`w(x) = 1`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(weight=lambda x: jnp.ones_like(x), domain=(-1, 1), scale=scale)

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.zeros_like(k)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, 2, 1 / (4 - 1 / jnp.where(k == 0, 1, k) ** 2))

    def _std_norm(self, k):
        return jnp.sqrt(2 / (2 * k + 1))

    def _std_scale(self, k):
        return jnp.exp(
            k * jnp.log(2) - gammaln(k + 1) + gammaln(k + 0.5) - gammaln(0.5)
        )


class ShiftedLegendre(ClassicalRecurrenceRelation):
    """Recurrence relation for Shifted Legendre Polynomials :math:`P^*_n(x)`

    Shifted Legendre polynomials are orthogonal on the interval (0, 1)
    with the weight function :math:`w(x) = 1`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(weight=lambda x: jnp.ones_like(x), domain=(0, 1), scale=scale)

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return 0.5 * jnp.ones_like(k)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, 1, 0.25 / (4 - 1 / jnp.where(k == 0, 1, k) ** 2))

    def _std_norm(self, k):
        return jnp.sqrt(1 / (2 * k + 1))

    def _std_scale(self, k):
        return jnp.exp(
            2 * k * jnp.log(2) - gammaln(k + 1) + gammaln(k + 0.5) - gammaln(0.5)
        )


class ChebyshevT(ClassicalRecurrenceRelation):
    """Recurrence relation for Chebyshev polynomials of the first kind :math:`T_n(x)`

    Chebyshev polynomials of the first kind are orthogonal on the interval (-1, 1)
    with the weight function :math:`w(x) = (1-x^2)^{-1/2}`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(
            weight=lambda x: 1.0 / jnp.sqrt(1 - x**2), domain=(-1, 1), scale=scale
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.zeros_like(k)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, jnp.pi, jnp.where(k == 1, 1 / 2, 1 / 4))

    def _std_norm(self, k):
        return jnp.sqrt(jnp.where(k == 0, jnp.pi, jnp.pi / 2))

    def _std_scale(self, k):
        return jnp.where(k == 0, 1, 2 ** jnp.maximum(0.0, k - 1.0))


class ChebyshevU(ClassicalRecurrenceRelation):
    """Recurrence relation for Chebyshev polynomials of the second kind :math:`U_n(x)`

    Chebyshev polynomials of the second kind are orthogonal on the interval (-1, 1)
    with the weight function :math:`w(x) = (1-x^2)^{1/2}`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(
            weight=lambda x: jnp.sqrt(1 - x**2), domain=(-1, 1), scale=scale
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.zeros_like(k)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, jnp.pi / 2, 1 / 4)

    def _std_norm(self, k):
        return jnp.full(k.shape, jnp.sqrt(jnp.pi / 2))

    def _std_scale(self, k):
        return 2 ** jnp.asarray(k).astype(float)


class ChebyshevV(ClassicalRecurrenceRelation):
    """Recurrence relation for Chebyshev polynomials of the third kind :math:`V_n(x)`

    Chebyshev polynomials of the third kind are orthogonal on the interval (-1, 1)
    with the weight function :math:`w(x) = (1-x)^{1/2} (1+x)^{-1/2}`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(
            weight=lambda x: jnp.sqrt((1 + x) / (1 - x)), domain=(-1, 1), scale=scale
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, 0.5, 0)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, jnp.pi, 1 / 4)

    def _std_norm(self, k):
        return jnp.full(k.shape, jnp.sqrt(jnp.pi))

    def _std_scale(self, k):
        return 2 ** jnp.asarray(k).astype(float)


class ChebyshevW(ClassicalRecurrenceRelation):
    """Recurrence relation for Chebyshev polynomials of the fourth kind :math:`W_n(x)`

    Chebyshev polynomials of the fourth kind are orthogonal on the interval (-1, 1)
    with the weight function :math:`w(x) = (1-x)^{-1/2} (1+x)^{1/2}`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(
            weight=lambda x: jnp.sqrt((1 - x) / (1 + x)), domain=(-1, 1), scale=scale
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, -0.5, 0)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, jnp.pi, 1 / 4)

    def _std_norm(self, k):
        return jnp.full(k.shape, jnp.sqrt(jnp.pi))

    def _std_scale(self, k):
        return 2 ** jnp.asarray(k).astype(float)


class Gegenbauer(ClassicalRecurrenceRelation):
    r"""Recurrence relation for Gegenbauer polynomials :math:`C^\lambda_n(x)`

    Also known as Ultraspherical harmonics.

    Gegenbauer polynomials are orthogonal on the interval (-1, 1)
    with the weight function :math:`w(x) = (1-x^2)^{\lambda - 1/2}`

    Parameters
    ----------
    lmbda : float > -1/2, != 0
        Hyperparameter λ.
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    lmbda: float

    def __init__(self, lmbda, scale="standard"):
        lmbda = eqx.error_if(lmbda, lmbda <= -0.5, "lmbda must be > -1/2")
        lam_zero_err = """
        Classical Gegenbauer polynomials are undefined for lmbda==0,
        consider using ChebyshevT which has similar orthogonality properties
        but with a well behaved normalization"""
        lmbda = eqx.error_if(lmbda, lmbda == 0.0, lam_zero_err)
        self.lmbda = lmbda
        super().__init__(
            weight=lambda x: (1 - x**2) ** (self.lmbda - 0.5),
            domain=(-1, 1),
            scale=scale,
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.zeros_like(k)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        b0 = jnp.sqrt(jnp.pi) * jnp.exp(
            gammaln(self.lmbda + 0.5) - gammaln(self.lmbda + 1)
        )
        bknum = jnp.where(self.lmbda == 0, 1, k * (k + 2 * self.lmbda - 1))
        bkden = jnp.where(
            self.lmbda == 0, 4, (4 * (k + self.lmbda) * (k + self.lmbda - 1))
        )
        return jnp.where(k == 0, b0, bknum / jnp.where(k == 0, 1, bkden))

    def _std_norm(self, k):
        lognum = (
            (1 - 2 * self.lmbda) * jnp.log(2)
            + jnp.log(jnp.pi)
            + gammaln(k + 2 * self.lmbda)
        )
        sgn = jnp.sign(k + self.lmbda) * jax.scipy.special.gammasgn(k + 2 * self.lmbda)
        logden = (
            jnp.log(jnp.abs(k + self.lmbda)) + 2 * gammaln(self.lmbda) + gammaln(k + 1)
        )
        return sgn * jnp.exp(0.5 * (lognum - logden))

    def _std_scale(self, k):
        return jnp.exp(
            k * jnp.log(2)
            - gammaln(k + 1)
            + gammaln(self.lmbda + k)
            - gammaln(self.lmbda)
        )


class Jacobi(ClassicalRecurrenceRelation):
    r"""Recurrence relation for Jacobi polynomials :math:`P^{(\alpha, \beta)}_n(x)`

    Jacobi polynomials are orthogonal on the interval (-1, 1)
    with the weight function :math:`w(x) = (1-x)^\alpha (1+x)^\beta`

    Parameters
    ----------
    alpha, beta : float > -1
        Hyperparameters α, β.
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    alpha: float
    beta: float

    def __init__(self, alpha, beta, scale="standard"):
        self.alpha = eqx.error_if(alpha, alpha <= -1, "alpha must be > -1")
        self.beta = eqx.error_if(beta, beta <= -1, "beta must be > -1")
        super().__init__(
            weight=lambda x: (1 - x) ** self.alpha * (1 + x) ** self.beta,
            domain=(-1, 1),
            scale=scale,
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        a, b = self.alpha, self.beta
        num = jnp.where(k == 0, b - a, b**2 - a**2)
        den = jnp.where(k == 0, a + b + 2, (2 * k + a + b) * (2 * k + a + b + 2))
        return num / den

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        a, b = self.alpha, self.beta
        b0 = jnp.exp(
            (a + b + 1) * jnp.log(2)
            + gammaln(a + 1)
            + gammaln(b + 1)
            - gammaln(a + b + 1)
        ) / (a + b + 1)
        num = jnp.where(
            k > 1, 4 * k * (k + a) * (k + b) * (k + a + b), 4 * k * (k + a) * (k + b)
        )
        den = jnp.where(
            k > 1,
            (2 * k + a + b) ** 2 * (2 * k + a + b + 1) * (2 * k + a + b - 1),
            (2 * k + a + b) ** 2 * (2 * k + a + b + 1),
        )
        return jnp.where(k == 0, b0, num / den)

    def _std_norm(self, k):
        a, b = self.alpha, self.beta
        lognum = (a + b + 1) * jnp.log(2) + gammaln(k + a + 1) + gammaln(k + b + 1)
        logden = jnp.log(2 * k + a + b + 1) + gammaln(k + a + b + 1) + gammaln(k + 1)
        return jnp.exp(0.5 * (lognum - logden))

    def _std_scale(self, k):
        a, b = self.alpha, self.beta
        logm = (
            gammaln(2 * k + a + b + 1)
            - k * jnp.log(2)
            - gammaln(k + 1)
            - gammaln(k + a + b + 1)
        )
        return jnp.exp(logm)


class Laguerre(ClassicalRecurrenceRelation):
    """Recurrence relation for Laguerre polynomials :math:`L_n(x)`

    Laguerre polynomials are orthogonal on the interval (0, inf)
    with the weight function :math:`w(x) = e^{-x}`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(weight=lambda x: jnp.exp(-x), domain=(0, jnp.inf), scale=scale)

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return 2 * k + 1

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, 1, k**2)

    def _std_norm(self, k):
        return jnp.ones(jnp.asarray(k).shape)

    def _std_scale(self, k):
        return (-1) ** k * jnp.exp(-gammaln(k + 1))


class GeneralizedLaguerre(ClassicalRecurrenceRelation):
    r"""Recurrence relation for Generalized Laguerre polynomials :math:`L^\alpha_n(x)`

    Generalized Laguerre polynomials are orthogonal on the interval (0, inf)
    with the weight function :math:`w(x) = x^\alpha e^{-x}`

    Parameters
    ----------
    alpha : float > -1
        Hyperparameter α.
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    alpha: float

    def __init__(self, alpha, scale="standard"):
        self.alpha = eqx.error_if(alpha, alpha <= -1, "alpha must be > -1")
        super().__init__(
            weight=lambda x: x**self.alpha * jnp.exp(-x),
            domain=(0, jnp.inf),
            scale=scale,
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return 2 * k + self.alpha + 1

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, jnp.exp(gammaln(self.alpha + 1)), k * (k + self.alpha))

    def _std_norm(self, k):
        return jnp.exp(-0.5 * gammaln(k + 1) + 0.5 * gammaln(k + self.alpha + 1))

    def _std_scale(self, k):
        return (-1) ** k * jnp.exp(-gammaln(k + 1))


class Hermite(ClassicalRecurrenceRelation):
    """Recurrence relation for (physicists) Hermite polynomials :math:`H_n(x)`

    Hermite polynomials are orthogonal on the interval (-inf, inf)
    with the weight function :math:`w(x) = e^{-x^2}`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(
            weight=lambda x: jnp.exp(-(x**2)), domain=(-jnp.inf, jnp.inf), scale=scale
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.zeros_like(k)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, jnp.sqrt(jnp.pi), k / 2)

    def _std_norm(self, k):
        return jnp.sqrt(jnp.sqrt(jnp.pi)) * 2 ** (k / 2) * jnp.exp(gammaln(k + 1) / 2)

    def _std_scale(self, k):
        return 2 ** jnp.asarray(k).astype(float)


class HermiteE(ClassicalRecurrenceRelation):
    """Recurrence relation for (probabalists) Hermite polynomials :math:`He_n(x)`

    Hermite polynomials are orthogonal on the interval (-inf, inf)
    with the weight function :math:`w(x) = e^{-x^2/2}`

    Parameters
    ----------
    scale : {"standard", "monic", "normalized"}
        "standard" corresponds to the common scaling found in textbooks such as
        Abramowitz & Stegun. "monic" scales them such that the leading coefficient is 1.
        "normalized" scales them to have a weighted norm of 1.
    """

    def __init__(self, scale="standard"):
        super().__init__(
            weight=lambda x: jnp.exp(-0.5 * x**2),
            domain=(-jnp.inf, jnp.inf),
            scale=scale,
        )

    def a(self, k):
        """`a` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.zeros_like(k)

    def b(self, k):
        """`b` coefficients of the monic three term recurrence relation."""
        k = _asarray(k)
        return jnp.where(k == 0, jnp.sqrt(2 * jnp.pi), k)

    def _std_norm(self, k):
        return jnp.sqrt(jnp.sqrt(2 * jnp.pi)) * jnp.exp(gammaln(k + 1) / 2)

    def _std_scale(self, k):
        return jnp.ones_like(k)


@jax.jit
def _polyval(x, n, a, b):
    a, b = map(lambda x: jnp.atleast_1d(jnp.asarray(x)), (a, b))
    x = jnp.asarray(x)

    p0 = jnp.zeros_like(x)
    p1 = jnp.ones_like(x)
    pn = p1

    def body(i, state):
        p0, p1, pn = state
        pn = (x - a[i]) * p1 - b[i] * p0
        p0 = p1
        p1 = pn
        return p0, p1, pn

    npos = lambda: jax.lax.fori_loop(0, n, body, (p0, p1, pn))[-1]
    nneg = lambda: jnp.zeros_like(x)
    return jax.lax.cond(n >= 0, npos, nneg)


def generate_recurrence(weight, domain, n, scale="monic", quadrule=None, quadopts=None):
    r"""Generate recurrence relation coefficients for orthogonal polynomial family.

    Finds coefficients :math:`a_i, b_i, g_i` such that

    .. math::

        \begin{align}
        p_{-1}(x) &= 0 \\
        p_0(x) &= 1 \\
        p_{i+1}(x) &= (x-a_i) p_i(x) - b_i p_{i-1}(x) \\
        \int_D p_i(x) p_j(x) w(x) dx &= g^2_i \delta_{i,j}
        \end{align}


    Parameters
    ----------
    weight : callable
        Weight function.
    domain : tuple of float
        Lower and upper bounds for the domain of the polynomials.
    n : int
        Number of terms to generate, ie, highest order of polynomial desired.
    scale : {"monic", "normalized"}
        How to scale the resulting polynomials.
    quadrule : quadax.AbstractQuadratureRule, optional
        Quadrature rule to use for computing integrals in generating recurrence
        coefficients. Defaults to ``quadax.TanhSinhRule(order=129)``.
    quadopts : dict, optional
        Additional options passed to ``quadax.adaptive_quadrature``. Default options
        are ``epsabs=1e-15``, ``epsrel=1e-15``, ``max_ninter=500``.

    Returns
    -------
    rec : TabulatedRecurrenceRelation
        Recurrence relation coefficients and polynomial norms.

    Notes
    -----
    Requires the ``quadax`` package to be installed.

    """
    assert scale in ["monic", "normalized"]
    try:
        import quadax
    except ImportError as e:
        raise ImportError(
            "quadax must be installed (use ``pip install quadax``) "
            + "to generate custom orthogonal polynomials."
        ) from e

    # p-adaptive might be better here, or bootstrapped gauss
    rule = quadrule or quadax.TanhSinhRule(129)
    opts = quadopts or {}
    opts.setdefault("epsabs", 1e-15)
    opts.setdefault("epsrel", 1e-15)
    opts.setdefault("max_ninter", 500)
    opts.setdefault("interval", jnp.asarray(domain))
    quad = lambda fun: quadax.adaptive_quadrature(rule, fun, **opts)

    @jax.jit
    def inner(n, a, b):
        fun = lambda x: _polyval(x, n, a, b) ** 2 * weight(x)
        return quad(fun)

    @jax.jit
    def innerx(n, a, b):
        fun = lambda x: x * _polyval(x, n, a, b) ** 2 * weight(x)
        return quad(fun)

    def body(i, state):
        aa, bb, cc, errs, status = state
        m0, out = inner(i, aa, bb)
        errs = errs.at[i, 0].set(out.err)
        status = status.at[i, 0].set(out.status)
        cc = cc.at[i].set(m0)
        ai, out = innerx(i, aa, bb)
        errs = errs.at[i, 1].set(out.err)
        status = status.at[i, 1].set(out.status)
        aa = aa.at[i].set(ai / cc[i])
        bb = bb.at[i].set(jnp.where(i == 0, m0, cc[i] / cc[i - 1]))
        return aa, bb, cc, errs, status

    aa = jnp.zeros(n)
    bb = jnp.zeros(n)
    cc = jnp.zeros(n)
    status = jnp.zeros((n, 2))
    errs = jnp.zeros((n, 2))

    aa, bb, cc, errs, status = jax.lax.fori_loop(0, n, body, (aa, bb, cc, errs, status))
    # TODO: figure out uncertainties better
    g = jnp.sqrt(cc)
    if scale == "monic":
        m = jnp.ones_like(g)
    else:  # normalized
        m = 1 / g

    return TabulatedRecurrenceRelation(weight, domain, aa, bb, g, m)
